"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.promiseTriggersDown = exports.promiseTriggersUp = exports.generateDown = exports.generateUp = void 0;
const sql_1 = require("@deep-foundation/hasura/sql");
const generateUp = (options) => () => __awaiter(void 0, void 0, void 0, function* () {
    const { schemaName, tableName, valueType, customColumnsSql = '', customAfterSql = '', linkRelation, linksTableName, api, deep } = options;
    yield api.sql((0, sql_1.sql) `
    CREATE TABLE ${schemaName}."${tableName}" (id bigint PRIMARY KEY, link_id bigint, ${customColumnsSql ? customColumnsSql : `value ${valueType}`});
    CREATE SEQUENCE ${tableName}_id_seq
    AS bigint START WITH 1 INCREMENT BY 1 NO MINVALUE NO MAXVALUE CACHE 1;
    ALTER SEQUENCE ${tableName}_id_seq OWNED BY ${schemaName}."${tableName}".id;
    ALTER TABLE ONLY ${schemaName}."${tableName}" ALTER COLUMN id SET DEFAULT nextval('${tableName}_id_seq'::regclass);
    CREATE INDEX IF NOT EXISTS ${tableName}__id_hash ON ${tableName} USING hash (id);
    CREATE INDEX IF NOT EXISTS ${tableName}__link_id_hash ON ${tableName} USING hash (link_id);
    CREATE INDEX IF NOT EXISTS ${tableName}__link_id_btree ON ${tableName} USING btree (link_id);
    ${customColumnsSql ? '' : `CREATE INDEX IF NOT EXISTS ${tableName}__value_btree ON ${tableName} USING btree (value);`}
    ${customAfterSql}
  `);
    yield api.query({
        type: 'track_table',
        args: {
            schema: schemaName,
            name: tableName,
        },
    });
    if (linkRelation && linksTableName) {
        yield api.query({
            type: 'create_object_relationship',
            args: {
                table: tableName,
                name: 'link',
                type: 'one_to_one',
                using: {
                    manual_configuration: {
                        remote_table: {
                            schema: schemaName,
                            name: linksTableName,
                        },
                        column_mapping: {
                            link_id: 'id',
                        },
                        insertion_order: 'after_parent',
                    },
                },
            },
        });
        yield api.query({
            type: 'create_object_relationship',
            args: {
                table: linksTableName,
                name: linkRelation,
                type: 'one_to_one',
                using: {
                    manual_configuration: {
                        remote_table: {
                            schema: schemaName,
                            name: tableName,
                        },
                        column_mapping: {
                            id: 'link_id',
                        },
                        insertion_order: 'after_parent',
                    },
                },
            },
        });
        yield api.query({
            type: 'create_event_trigger',
            args: {
                name: tableName,
                table: tableName,
                webhook: `${process.env.MIGRATIONS_DEEPLINKS_URL}/api/values`,
                insert: {
                    columns: "*",
                    payload: '*',
                },
                update: {
                    columns: '*',
                    payload: '*',
                },
                delete: {
                    columns: '*'
                },
                replace: false,
            },
        });
    }
});
exports.generateUp = generateUp;
const generateDown = (options) => () => __awaiter(void 0, void 0, void 0, function* () {
    const { schemaName, tableName, linkRelation, linksTableName, api, deep } = options;
    if (linkRelation && linksTableName)
        yield api.query({
            type: 'drop_relationship',
            args: {
                table: linksTableName,
                relationship: linkRelation,
            },
        });
    if (linkRelation && linksTableName)
        yield api.query({
            type: 'drop_relationship',
            args: {
                table: tableName,
                relationship: 'link',
            },
        });
    yield api.query({
        type: 'untrack_table',
        args: {
            table: {
                schema: schemaName,
                name: tableName,
            },
        },
    });
    yield api.sql((0, sql_1.sql) `
    DROP TABLE ${schemaName}."${tableName}";
  `);
});
exports.generateDown = generateDown;
const promiseTriggersUp = (options) => () => __awaiter(void 0, void 0, void 0, function* () {
    const { schemaName, tableName, valueType, customColumnsSql = '', customAfterSql = '', linkRelation, linksTableName, api, deep } = options;
    const promiseTypeId = yield deep.id('@deep-foundation/core', 'Promise');
    const thenTypeId = yield deep.id('@deep-foundation/core', 'Then');
    const handleUpdateTypeId = yield deep.id('@deep-foundation/core', 'HandleUpdate');
    yield api.sql((0, sql_1.sql) `CREATE OR REPLACE FUNCTION ${tableName}__promise__insert__function() RETURNS TRIGGER AS $trigger$ DECLARE PROMISE bigint;
  BEGIN
    IF (
        EXISTS(
          SELECT 1
          FROM links handle_update, links updated_link
          WHERE
              updated_link.id = NEW."link_id"
          AND handle_update.from_id = updated_link."type_id"
          AND handle_update.type_id = ${handleUpdateTypeId}
        )
    ) THEN
      INSERT INTO links ("type_id") VALUES (${promiseTypeId}) RETURNING id INTO PROMISE;
      INSERT INTO links ("type_id","from_id","to_id") VALUES (${thenTypeId},NEW."link_id",PROMISE);
    END IF;
    RETURN NEW;
  END; $trigger$ LANGUAGE plpgsql;`);
    yield api.sql((0, sql_1.sql) `CREATE TRIGGER ${tableName}__promise__insert__trigger AFTER INSERT ON "${tableName}" FOR EACH ROW EXECUTE PROCEDURE ${tableName}__promise__insert__function();`);
    yield api.sql((0, sql_1.sql) `CREATE OR REPLACE FUNCTION ${tableName}__promise__update__function() RETURNS TRIGGER AS $trigger$ DECLARE PROMISE bigint;
  BEGIN
    IF (
        EXISTS(
          SELECT 1
          FROM links handle_update, links updated_link
          WHERE
              updated_link.id = NEW."link_id"
          AND handle_update.from_id = updated_link."type_id"
          AND handle_update.type_id = ${handleUpdateTypeId}
        )
    ) THEN
      INSERT INTO links ("type_id") VALUES (${promiseTypeId}) RETURNING id INTO PROMISE;
      INSERT INTO links ("type_id","from_id","to_id") VALUES (${thenTypeId},NEW."link_id",PROMISE);
    END IF;
    RETURN NEW;
  END; $trigger$ LANGUAGE plpgsql;`);
    yield api.sql((0, sql_1.sql) `CREATE TRIGGER ${tableName}__promise__update__trigger AFTER INSERT ON "${tableName}" FOR EACH ROW EXECUTE PROCEDURE ${tableName}__promise__update__function();`);
    yield api.sql((0, sql_1.sql) `CREATE OR REPLACE FUNCTION ${tableName}__promise__delete__function() RETURNS TRIGGER AS $trigger$ DECLARE PROMISE bigint;
  BEGIN
    IF (
        EXISTS(
          SELECT 1
          FROM links handle_update, links updated_link
          WHERE
              updated_link.id = OLD."link_id"
          AND handle_update.from_id = updated_link."type_id"
          AND handle_update.type_id = ${handleUpdateTypeId}
        )
    ) THEN
      INSERT INTO links ("type_id") VALUES (${promiseTypeId}) RETURNING id INTO PROMISE;
      INSERT INTO links ("type_id","from_id","to_id") VALUES (${thenTypeId},OLD."link_id",PROMISE);
    END IF;
    RETURN OLD;
  END; $trigger$ LANGUAGE plpgsql;`);
    yield api.sql((0, sql_1.sql) `CREATE TRIGGER ${tableName}__promise__delete__trigger BEFORE DELETE ON "${tableName}" FOR EACH ROW EXECUTE PROCEDURE ${tableName}__promise__delete__function();`);
});
exports.promiseTriggersUp = promiseTriggersUp;
const promiseTriggersDown = (options) => () => __awaiter(void 0, void 0, void 0, function* () {
    const { schemaName, tableName, linkRelation, linksTableName, api, deep } = options;
    yield api.sql((0, sql_1.sql) `DROP TRIGGER IF EXISTS ${tableName}__promise__insert__trigger ON "${tableName}";`);
    yield api.sql((0, sql_1.sql) `DROP FUNCTION IF EXISTS ${tableName}__promise__insert__function CASCADE;`);
    yield api.sql((0, sql_1.sql) `DROP TRIGGER IF EXISTS ${tableName}__promise__update__trigger ON "${tableName}";`);
    yield api.sql((0, sql_1.sql) `DROP FUNCTION IF EXISTS ${tableName}__promise__update__function CASCADE;`);
    yield api.sql((0, sql_1.sql) `DROP TRIGGER IF EXISTS ${tableName}__promise__delete__trigger ON "${tableName}";`);
    yield api.sql((0, sql_1.sql) `DROP FUNCTION IF EXISTS ${tableName}__promise__delete__function CASCADE;`);
});
exports.promiseTriggersDown = promiseTriggersDown;
//# sourceMappingURL=type-table.js.map